from collections import deque
from contextlib import suppress
from typing import Any, Optional, TypeVar, Union

_U = TypeVar('_U')


def setup_type(idl_type: Any, type_: str) -> None:
    todo = deque([idl_type])

    with suppress(IndexError):
        while idl_type := todo.popleft():
            idl_type.type = type_

            if idl_type.generic:
                todo.extend(idl_type.idl_type)


def escaped_name(string: Optional[_U]) -> Union[Optional[_U], str]:
    if isinstance(string, str):
        return string.lstrip('_')

    return string



from typing import NamedTuple

from .nodes import Ast


class Variable(NamedTuple):

    type: str
    ast: Ast


def create_js_variable(type_: str, ast: Ast) -> Variable:
    return Variable(type=type_, ast=ast)



from typing import Type

from .js.statements import (
    create_attribute,
    create_object,
    create_operation,
    create_dict,
    create_property,
)
from .js.variable import Variable as JsVariable, create_js_variable
from .visitor import Visitor
from .utils import unique_name
from .webidl.nodes import (
    Ast as WebIDLAst,
    Attribute,
    Constructor,
    Definition,
    Dictionary,
    Interface,
    Operation,
)


class Converter(Visitor[WebIDLAst]):

    def __init__(self, builder):
        self.variables: list[JsVariable] = []
        self.dependencies: list[Converter] = []

        self._builder = builder

    def _calculate_node(self, node: WebIDLAst):
        dependency, property_ = self._builder.create(node)

        if dependency is not None:
            self.dependencies.append(dependency)

        return property_


class DictionaryConverter(Converter):

    def visit_dictionary(self, node: Dictionary) -> None:
        self.variables.append(
            create_js_variable(
                type_=node.name,
                ast=create_dict(
                    name=unique_name(),
                    properties=[
                        create_property(
                            key=member.name,  # type: ignore
                            value=self._calculate_node(node=member),
                        )
                        for member in node.members
                    ]
                )
            )
        )


class InterfaceConverter(Converter):

    def __init__(self, builder):
        super().__init__(builder=builder)

        self._type = None
        self._name = None

    def visit_interface(self, node: Interface) -> None:
        if node.partial is True:
            return

        self._type = node.name
        self._name = unique_name()

        self.generic_visit(node)

    def visit_constructor(self, node: Constructor) -> None:
        self.variables.append(
            create_js_variable(
                type_=self._type,
                ast=create_object(
                    name=self._name,
                    progenitor=self._type,
                    arguments=[
                        self._calculate_node(node=argument)
                        for argument in node.arguments
                    ],
                ),
            )
        )

        self.generic_visit(node)

    def visit_attribute(self, node: Attribute) -> None:
        self.variables.append(
            create_js_variable(
                type_=node.idl_type.idl_type,  # type: ignore
                ast=create_attribute(
                    name=unique_name(),
                    progenitor=self._name,
                    method=node.name
                ),
            )
        )

        self.generic_visit(node)

    def visit_operation(self, node: Operation) -> None:
        self.variables.append(
            create_js_variable(
                type_=node.idl_type.idl_type,  # type: ignore
                ast=create_operation(
                    name=unique_name(),
                    progenitor=self._name,
                    method=node.name,
                    arguments=[
                        self._calculate_node(node=argument)
                        for argument in node.arguments
                    ],
                )
            )
        )

        self.generic_visit(node)


CONVERTER_MAP: dict[str, Type[Converter]] = {
    'dictionary': DictionaryConverter,
    'interface': InterfaceConverter,
}


def convert(builder, definition: Definition) -> Converter:
    return (
        (converter := CONVERTER_MAP[definition.type](builder=builder)).visit(definition) or
        converter
    )



import logging
from logging.config import dictConfig as logging_dict_config

import click


TRACE_LOG_LEVEL = 5


LOG_LEVELS = {
    'critical': logging.CRITICAL,
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'info': logging.INFO,
    'debug': logging.DEBUG,
    'trace': TRACE_LOG_LEVEL,
}


LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            '()': 'idl2js.logger.IDL2JSFormatter',
            'fmt': '%(asctime)s %(levelname)s %(message)s',
            'datefmt': '%H:%M:%S',
        },
    },
    'handlers': {
        'default': {
            'formatter': 'default',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stderr',
        },
    },
    'loggers': {
        'idl2js': {'handlers': ['default'], 'level': 'ERROR'},
    },
}


class IDL2JSFormatter(logging.Formatter):

    level_name_colors = {
        TRACE_LOG_LEVEL: lambda message: click.style(message, fg='blue'),
        logging.DEBUG: lambda message: click.style(message, fg='cyan'),
        logging.INFO: lambda message: click.style(message, fg='green'),
        logging.WARNING: lambda message: click.style(message, fg='yellow'),
        logging.ERROR: lambda message: click.style(message, fg='red'),
        logging.CRITICAL: lambda message: click.style(message, fg='bright_red'),
    }

    def formatMessage(self, record):
        colored = self.level_name_colors[record.levelno]

        record.__dict__['message'] = colored(record.msg % record.args)
        record.__dict__['asctime'] = colored(f'[+] {record.asctime}')
        record.__dict__['levelname'] = colored(f'{record.levelname}:')

        return super().formatMessage(record)


def configure_logging(log_level):
    logging.addLevelName(TRACE_LOG_LEVEL, 'TRACE')
    logging_dict_config(LOGGING_CONFIG)
    logging.getLogger('idl2js').setLevel(LOG_LEVELS[log_level])



from operator import attrgetter
from typing import Any, Generic, TypeVar

import stringcase
from attr import fields


AstType = TypeVar('AstType')


class Visitor(Generic[AstType]):

    def visit(self, node: AstType) -> Any:
        """Visit a node."""
        visitor = getattr(
            self,
            f'visit_{stringcase.snakecase(node.__class__.__name__)}',
            self.generic_visit,
        )

        return visitor(node)

    def generic_visit(self, node: AstType) -> None:
        """Called if no explicit visitor function exists for a node."""
        ast_type = self.__orig_bases__[0].__args__  # type: ignore

        for name in map(attrgetter('name'), fields(type(node))):
            field = getattr(node, name)

            if isinstance(field, ast_type):
                self.visit(field)
            elif isinstance(field, list):
                for item in field:
                    if isinstance(item, ast_type):
                        self.visit(item)



from functools import cached_property
from operator import attrgetter
from collections import deque

from more_itertools import flatten, first_true, first

from .built_in_types import BuiltInTypes
from .converter import convert
from .js.statements import create_literal, create_identifier


class DefinitionStorage:

    def __init__(self, idls):
        self._definitions = list(flatten(map(attrgetter('definitions'), idls)))

    @cached_property
    def build_definition(self):
        return [
            definition
            for definition in self._definitions
            if definition.type == 'interface' and definition.partial is False
        ]

    def find_by_type(self, idl_type):
        return first_true(self._definitions, pred=lambda definition: definition.name == idl_type)


class Builder:

    def __init__(self, std_types: BuiltInTypes, definition_storage: DefinitionStorage):
        self._std_types = std_types
        self._definition_storage = definition_storage

    def create(self, node):
        if node.idl_type.idl_type in self._std_types:
            return None, create_literal(self._std_types.generate(node.idl_type.idl_type))

        definition = self._definition_storage.find_by_type(node.idl_type.idl_type)
        converter = convert(builder=self, definition=definition)

        return converter, create_identifier(first(converter.variables).ast.expression.left.name)


def build(definition_storage, builder):
    for definition in definition_storage.build_definition:
        result = []

        todo = deque([convert(builder=builder, definition=definition)])
        while todo:
            item = todo.popleft()
            result.append(item.variables)
            todo.extendleft(item.dependencies)

        yield from flatten(reversed(result))



from collections import defaultdict


class Storage:

    def __init__(self):
        self.variables = []
        self._by_type = defaultdict(list)

    def add(self, variable):
        self.variables.append(variable)

    def __setitem__(self, key, value):
        self._by_type[key].append(value)

    def __getitem__(self, item):
        return self._by_type[item]



from .parser import Parser as WebIDLParser
from .visitor import Visitor as WebIDLVisitor
from .webidl import parse, parse_as_dict, validate


__all__ = (
    'WebIDLParser',
    'WebIDLVisitor',
    'parse',
    'parse_as_dict',
    'validate',
)



import os
from collections.abc import Mapping
from typing import Any, Iterator


class Block(Mapping):

    def __init__(self, /, **kwargs: Any):
        self._data = kwargs

    def __getitem__(self, item: str) -> str:
        return self._data[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)


class Config:

    def __init__(self):
        self._env = os.environ.copy()
        self._config = {}

    def _setup_parser(self):
        self._config['PARSER'] = Block(
            **self._parser_defaults(),
        )

    @staticmethod
    def _parser_defaults():
        return Block(
            num_of_process=4,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['idl2js', 'idl2js.js', 'idl2js.webidl', 'idl2js.webidl.generated']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime>=4.9.2,<5.0.0',
 'attrs>=19.3.0,<20.0.0',
 'click>=7.1.2,<8.0.0',
 'graphviz>=0.16,<0.17',
 'more_itertools>=8.4.0,<9.0.0',
 'stringcase>=1.2.0,<2.0.0']

setup_kwargs = {
    'name': 'idl2js',
    'version': '0.1.0',
    'description': 'Grammar-based Fuzzer that uses WebIDL as a grammar.',
    'long_description': '# idl2js\n\n[![Build Status](https://img.shields.io/travis/PrVrSs/idl2js/master?style=plastic)](https://travis-ci.org/github/PrVrSs/idl2js)\n[![Codecov](https://img.shields.io/codecov/c/github/PrVrSs/idl2js?style=plastic)](https://codecov.io/gh/PrVrSs/idl2js)\n[![Python Version](https://img.shields.io/badge/python-3.9-blue?style=plastic)](https://www.python.org/)\n[![License](https://img.shields.io/cocoapods/l/A?style=plastic)](https://github.com/PrVrSs/idl2js/blob/master/LICENSE)\n\n\n## Quick start\n\n```shell script\npip install idl2js\n```\n\n\n### Links\n\n\n* [searchfox - webidl](https://searchfox.org/mozilla-central/source/dom/webidl)\n* [original webidl parser](https://github.com/w3c/webidl2.js)\n* [TSJS-lib-generator](https://github.com/microsoft/TSJS-lib-generator/tree/master/inputfiles/idl)\n* [ECMAScript® 2021 Language Specification](https://tc39.es/ecma262/)\n\n\n## Contributing\n\nAny help is welcome and appreciated.\n\n\n## License\n\n*idl2js* is licensed under the terms of the Apache-2.0 License (see the file LICENSE).\n',
    'author': 'Sergey Reshetnikov',
    'author_email': 'resh.sersh@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PrVrSs/idl2js',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Generated from /home/sergey/PycharmProjects/idl2js/grammar/WebIDLParser.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .WebIDLParser import WebIDLParser
else:
    from WebIDLParser import WebIDLParser

# This class defines a complete generic visitor for a parse tree produced by WebIDLParser.

class WebIDLParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by WebIDLParser#webIDL.
    def visitWebIDL(self, ctx:WebIDLParser.WebIDLContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#definitions.
    def visitDefinitions(self, ctx:WebIDLParser.DefinitionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#extendedDefinition.
    def visitExtendedDefinition(self, ctx:WebIDLParser.ExtendedDefinitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#definition.
    def visitDefinition(self, ctx:WebIDLParser.DefinitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#callbackOrInterfaceOrMixin.
    def visitCallbackOrInterfaceOrMixin(self, ctx:WebIDLParser.CallbackOrInterfaceOrMixinContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#interfaceOrMixin.
    def visitInterfaceOrMixin(self, ctx:WebIDLParser.InterfaceOrMixinContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#interfaceRest.
    def visitInterfaceRest(self, ctx:WebIDLParser.InterfaceRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#partial.
    def visitPartial(self, ctx:WebIDLParser.PartialContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#partialInterfaceOrPartialMixin.
    def visitPartialInterfaceOrPartialMixin(self, ctx:WebIDLParser.PartialInterfaceOrPartialMixinContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#partialInterfaceRest.
    def visitPartialInterfaceRest(self, ctx:WebIDLParser.PartialInterfaceRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#interfaceMembers.
    def visitInterfaceMembers(self, ctx:WebIDLParser.InterfaceMembersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#interfaceMember.
    def visitInterfaceMember(self, ctx:WebIDLParser.InterfaceMemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#partialInterfaceMembers.
    def visitPartialInterfaceMembers(self, ctx:WebIDLParser.PartialInterfaceMembersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#partialInterfaceMember.
    def visitPartialInterfaceMember(self, ctx:WebIDLParser.PartialInterfaceMemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#inheritance.
    def visitInheritance(self, ctx:WebIDLParser.InheritanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#mixinRest.
    def visitMixinRest(self, ctx:WebIDLParser.MixinRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#mixinMembers.
    def visitMixinMembers(self, ctx:WebIDLParser.MixinMembersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#mixinMember.
    def visitMixinMember(self, ctx:WebIDLParser.MixinMemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#includesStatement.
    def visitIncludesStatement(self, ctx:WebIDLParser.IncludesStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#callbackRestOrInterface.
    def visitCallbackRestOrInterface(self, ctx:WebIDLParser.CallbackRestOrInterfaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#callbackInterfaceMembers.
    def visitCallbackInterfaceMembers(self, ctx:WebIDLParser.CallbackInterfaceMembersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#callbackInterfaceMember.
    def visitCallbackInterfaceMember(self, ctx:WebIDLParser.CallbackInterfaceMemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#const_.
    def visitConst_(self, ctx:WebIDLParser.Const_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#constValue.
    def visitConstValue(self, ctx:WebIDLParser.ConstValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#readOnlyMember.
    def visitReadOnlyMember(self, ctx:WebIDLParser.ReadOnlyMemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#readOnlyMemberRest.
    def visitReadOnlyMemberRest(self, ctx:WebIDLParser.ReadOnlyMemberRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#readWriteAttribute.
    def visitReadWriteAttribute(self, ctx:WebIDLParser.ReadWriteAttributeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#inheritAttribute.
    def visitInheritAttribute(self, ctx:WebIDLParser.InheritAttributeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#attributeRest.
    def visitAttributeRest(self, ctx:WebIDLParser.AttributeRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#attributeName.
    def visitAttributeName(self, ctx:WebIDLParser.AttributeNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#defaultValue.
    def visitDefaultValue(self, ctx:WebIDLParser.DefaultValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#operation.
    def visitOperation(self, ctx:WebIDLParser.OperationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#regularOperation.
    def visitRegularOperation(self, ctx:WebIDLParser.RegularOperationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#operationRest.
    def visitOperationRest(self, ctx:WebIDLParser.OperationRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#operationName.
    def visitOperationName(self, ctx:WebIDLParser.OperationNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#operationNameKeyword.
    def visitOperationNameKeyword(self, ctx:WebIDLParser.OperationNameKeywordContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#argumentList.
    def visitArgumentList(self, ctx:WebIDLParser.ArgumentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#argument.
    def visitArgument(self, ctx:WebIDLParser.ArgumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#argumentRest.
    def visitArgumentRest(self, ctx:WebIDLParser.ArgumentRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#argumentName.
    def visitArgumentName(self, ctx:WebIDLParser.ArgumentNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#constructor.
    def visitConstructor(self, ctx:WebIDLParser.ConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#stringifier.
    def visitStringifier(self, ctx:WebIDLParser.StringifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#stringifierRest.
    def visitStringifierRest(self, ctx:WebIDLParser.StringifierRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#staticMember.
    def visitStaticMember(self, ctx:WebIDLParser.StaticMemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#staticMemberRest.
    def visitStaticMemberRest(self, ctx:WebIDLParser.StaticMemberRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#iterable.
    def visitIterable(self, ctx:WebIDLParser.IterableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#optionalArgumentList.
    def visitOptionalArgumentList(self, ctx:WebIDLParser.OptionalArgumentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#readWriteMaplike.
    def visitReadWriteMaplike(self, ctx:WebIDLParser.ReadWriteMaplikeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#maplikeRest.
    def visitMaplikeRest(self, ctx:WebIDLParser.MaplikeRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#readWriteSetlike.
    def visitReadWriteSetlike(self, ctx:WebIDLParser.ReadWriteSetlikeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#setlikeRest.
    def visitSetlikeRest(self, ctx:WebIDLParser.SetlikeRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#namespace.
    def visitNamespace(self, ctx:WebIDLParser.NamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#namespaceMembers.
    def visitNamespaceMembers(self, ctx:WebIDLParser.NamespaceMembersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#namespaceMember.
    def visitNamespaceMember(self, ctx:WebIDLParser.NamespaceMemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#dictionary.
    def visitDictionary(self, ctx:WebIDLParser.DictionaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#dictionaryMembers.
    def visitDictionaryMembers(self, ctx:WebIDLParser.DictionaryMembersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#dictionaryMember.
    def visitDictionaryMember(self, ctx:WebIDLParser.DictionaryMemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#default_.
    def visitDefault_(self, ctx:WebIDLParser.Default_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#enum_.
    def visitEnum_(self, ctx:WebIDLParser.Enum_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#callbackRest.
    def visitCallbackRest(self, ctx:WebIDLParser.CallbackRestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#typedef.
    def visitTypedef(self, ctx:WebIDLParser.TypedefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#type_.
    def visitType_(self, ctx:WebIDLParser.Type_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#typeWithExtendedAttributes.
    def visitTypeWithExtendedAttributes(self, ctx:WebIDLParser.TypeWithExtendedAttributesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#extendedAttributeList.
    def visitExtendedAttributeList(self, ctx:WebIDLParser.ExtendedAttributeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#extendedAttributeNoArgs.
    def visitExtendedAttributeNoArgs(self, ctx:WebIDLParser.ExtendedAttributeNoArgsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#extendedAttributeIdentList.
    def visitExtendedAttributeIdentList(self, ctx:WebIDLParser.ExtendedAttributeIdentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#extendedAttributeNamedArgList.
    def visitExtendedAttributeNamedArgList(self, ctx:WebIDLParser.ExtendedAttributeNamedArgListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#extendedAttributeIdent.
    def visitExtendedAttributeIdent(self, ctx:WebIDLParser.ExtendedAttributeIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#extendedAttributeArgList.
    def visitExtendedAttributeArgList(self, ctx:WebIDLParser.ExtendedAttributeArgListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#identifierList.
    def visitIdentifierList(self, ctx:WebIDLParser.IdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#identifier.
    def visitIdentifier(self, ctx:WebIDLParser.IdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#returnType.
    def visitReturnType(self, ctx:WebIDLParser.ReturnTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#singleType.
    def visitSingleType(self, ctx:WebIDLParser.SingleTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#unionType.
    def visitUnionType(self, ctx:WebIDLParser.UnionTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#unionMemberType.
    def visitUnionMemberType(self, ctx:WebIDLParser.UnionMemberTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#genericType.
    def visitGenericType(self, ctx:WebIDLParser.GenericTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#distinguishableType.
    def visitDistinguishableType(self, ctx:WebIDLParser.DistinguishableTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#primitiveType.
    def visitPrimitiveType(self, ctx:WebIDLParser.PrimitiveTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#constType.
    def visitConstType(self, ctx:WebIDLParser.ConstTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#promiseType.
    def visitPromiseType(self, ctx:WebIDLParser.PromiseTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#recordType.
    def visitRecordType(self, ctx:WebIDLParser.RecordTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#unrestrictedFloatType.
    def visitUnrestrictedFloatType(self, ctx:WebIDLParser.UnrestrictedFloatTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#floatType.
    def visitFloatType(self, ctx:WebIDLParser.FloatTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#unsignedIntegerType.
    def visitUnsignedIntegerType(self, ctx:WebIDLParser.UnsignedIntegerTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#integerType.
    def visitIntegerType(self, ctx:WebIDLParser.IntegerTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#stringType.
    def visitStringType(self, ctx:WebIDLParser.StringTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#bufferRelatedType.
    def visitBufferRelatedType(self, ctx:WebIDLParser.BufferRelatedTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#booleanLiteral.
    def visitBooleanLiteral(self, ctx:WebIDLParser.BooleanLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#floatLiteral.
    def visitFloatLiteral(self, ctx:WebIDLParser.FloatLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#null_.
    def visitNull_(self, ctx:WebIDLParser.Null_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#other.
    def visitOther(self, ctx:WebIDLParser.OtherContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WebIDLParser#argumentNameKeyword.
    def visitArgumentNameKeyword(self, ctx:WebIDLParser.ArgumentNameKeywordContext):
        return self.visitChildren(ctx)



del WebIDLParser


import json
import uuid
from pathlib import Path

import attr


def dump_js(variables, file='tmp.json'):
    with open(file, 'w') as fp:
        json.dump(
            obj=[attr.asdict(variable) for variable in variables],
            fp=fp,
            indent=4,
        )


def interleave(iterable, func, separator):
    it = iter(iterable)

    try:
        func(next(it))
    except StopIteration:
        pass
    else:
        for item in it:
            separator()
            func(item)


def unique_name():
    return f'v_{uuid.uuid4().hex}'


def save(file_name, content):
    with open(Path(file_name), 'w') as file:
        file.write('\n'.join(content))



import abc
from typing import List, Optional, Union

import attr

from .utils import escaped_name


@attr.s
class Ast(abc.ABC):
    ...


@attr.s
class Definition(Ast):

    type: str = attr.ib()


@attr.s
class Member(Ast):
    ...


@attr.s
class Expression(Ast):
    ...


@attr.s
class Definitions(Ast):

    definitions: List[Definition] = attr.ib(factory=list)


@attr.s
class Null(Expression):

    type: str = attr.ib(default='null')


@attr.s
class NaN(Expression):

    type: str = attr.ib(default='NaN')


@attr.s
class Infinity(Expression):

    negative: bool = attr.ib()
    type: str = attr.ib(default='Infinity')


@attr.s
class Literal(Expression):

    type: Optional[str] = attr.ib()
    value: Union[str, list] = attr.ib()

    def __attrs_post_init__(self):
        self.value = escaped_name(self.value)


@attr.s
class Value(Expression):

    value: str = attr.ib()

    def __attrs_post_init__(self):
        self.value = escaped_name(self.value)


@attr.s
class LiteralList(Expression):

    type: str = attr.ib()
    value: List[Value] = attr.ib(factory=list)


@attr.s
class ExtendedAttribute(Ast):

    name: str = attr.ib()
    arguments: List[str] = attr.ib(factory=list)
    rhs: Optional[Union[Literal, LiteralList]] = attr.ib(default=None)
    type: str = attr.ib(default='extended-attribute')


@attr.s
class IdlType(Ast):

    idl_type: Union[List['IdlType'], 'IdlType', str] = attr.ib()
    type: Optional[str] = attr.ib(default=None)
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    nullable: bool = attr.ib(default=False)
    union: bool = attr.ib(default=False)
    generic: str = attr.ib(default='')

    def __attrs_post_init__(self):
        self.idl_type = escaped_name(self.idl_type)


@attr.s
class Argument(Ast):

    name: str = attr.ib()
    idl_type: IdlType = attr.ib()
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    default: Optional[Union[Literal, Value]] = attr.ib(default=None)
    type: str = attr.ib(default='argument')
    optional: bool = attr.ib(default=False)
    variadic: bool = attr.ib(default=False)

    def __attrs_post_init__(self):
        self.name = escaped_name(self.name)


@attr.s
class Operation(Member):

    name: str = attr.ib()
    idl_type: Optional[IdlType] = attr.ib()
    arguments: List[Argument] = attr.ib(factory=list)
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='operation')
    special: str = attr.ib(default='')


@attr.s
class Iterable_(Member):  # pylint: disable=invalid-name

    idl_type: List[IdlType] = attr.ib()
    arguments: List[Argument] = attr.ib()
    async_: bool = attr.ib()
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='iterable')
    readonly: bool = attr.ib(default=False)


@attr.s
class Attribute(Member):

    idl_type: IdlType = attr.ib()
    name: str = attr.ib()
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='attribute')
    readonly: bool = attr.ib(default=False)
    special: str = attr.ib(default='')

    def __attrs_post_init__(self):
        self.name = escaped_name(self.name)


@attr.s
class Const(Member):

    name: str = attr.ib()
    value: Expression = attr.ib()
    idl_type: IdlType = attr.ib()
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='const')


@attr.s
class Constructor(Member):

    arguments: List[Argument] = attr.ib(factory=list)
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='constructor')


@attr.s
class MapLike(Member):

    readonly: bool = attr.ib()
    idl_type: List[IdlType] = attr.ib()
    arguments: List[Argument] = attr.ib(factory=list)
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='maplike')
    async_: bool = attr.ib(default=False)


@attr.s
class SetLike(Member):

    readonly: bool = attr.ib()
    idl_type: List[IdlType] = attr.ib()
    arguments: List[Argument] = attr.ib(factory=list)
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='setlike')
    async_: bool = attr.ib(default=False)


@attr.s
class Field(Member):

    idl_type: IdlType = attr.ib()
    name: str = attr.ib()
    default: Optional[Literal] = attr.ib()
    required: bool = attr.ib(default=False)
    type: str = attr.ib(default='field')
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)


@attr.s
class CallbackInterface(Definition):

    name: str = attr.ib()
    members: List[Member] = attr.ib()
    inheritance: Optional[str] = attr.ib()
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='callback interface')
    partial: bool = attr.ib(default=False)


@attr.s
class Callback(Definition):

    name: str = attr.ib()
    idl_type: IdlType = attr.ib()
    arguments: List[Argument] = attr.ib(factory=list)
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='callback')


@attr.s
class Interface(Definition):

    members: List[Member] = attr.ib()
    name: str = attr.ib()
    inheritance: Optional[str] = attr.ib(default=None)
    partial: bool = attr.ib(default=False)
    type: str = attr.ib(default='interface')
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)

    def __attrs_post_init__(self):
        self.name = escaped_name(self.name)
        self.inheritance = escaped_name(self.inheritance)


@attr.s
class InterfaceMixin(Definition):

    members: List[Member] = attr.ib()
    name: str = attr.ib()
    inheritance: Optional[str] = attr.ib(default=None)
    partial: bool = attr.ib(default=False)
    type: str = attr.ib(default='interface mixin')
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)

    def __attrs_post_init__(self):
        self.name = escaped_name(self.name)
        self.inheritance = escaped_name(self.inheritance)


@attr.s
class Enum(Definition):

    name: str = attr.ib()
    type: str = attr.ib(default='enum')
    values: List[Literal] = attr.ib(factory=list)
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)


@attr.s
class Dictionary(Definition):

    name: str = attr.ib()
    inheritance: Optional[str] = attr.ib(default=None)
    members: List[Member] = attr.ib(factory=list)
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)
    type: str = attr.ib(default='dictionary')
    partial: bool = attr.ib(default=False)

    def __attrs_post_init__(self):
        self.name = escaped_name(self.name)


@attr.s
class Includes(Definition):

    target: str = attr.ib()
    includes: str = attr.ib()
    type: str = attr.ib(default='includes')
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)

    def __attrs_post_init__(self):
        self.target = escaped_name(self.target)
        self.includes = escaped_name(self.includes)


@attr.s
class Typedef(Definition):

    idl_type: IdlType = attr.ib()
    name: str = attr.ib()
    type: str = attr.ib(default='typedef')
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)

    def __attrs_post_init__(self):
        self.name = escaped_name(self.name)


@attr.s
class Namespace(Definition):

    members: List[Member] = attr.ib()
    name: str = attr.ib()
    inheritance: Optional[str] = attr.ib(default=None)
    partial: bool = attr.ib(default=False)
    type: str = attr.ib(default='namespace')
    ext_attrs: List[ExtendedAttribute] = attr.ib(factory=list)

    def __attrs_post_init__(self):
        self.name = escaped_name(self.name)



from functools import reduce, update_wrapper
from typing import ClassVar, Dict, List, NamedTuple, cast

from antlr4 import BailErrorStrategy, CommonTokenStream, FileStream
from antlr4.error.Errors import ParseCancellationException
from antlr4.error.ErrorListener import ErrorListener
from antlr4.error.ErrorStrategy import DefaultErrorStrategy, ErrorStrategy
from antlr4.Parser import ParserRuleContext

from .generated import WebIDLLexer, WebIDLParser

from ..error import IDLParseError


class SyntaxErrorInfo(NamedTuple):

    line: int
    column: int
    message: str

    def __repr__(self):  # pragma: no cover
        return f'{self.line}:{self.column} {self.message}'


class WebIDLErrorListener(ErrorListener):

    def __init__(self):
        self._errors: List[SyntaxErrorInfo] = []

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        self._errors.append(SyntaxErrorInfo(line, column, msg))

    @property
    def errors(self) -> List[SyntaxErrorInfo]:
        return self._errors


def _setup_strategy(strategy):
    def decorating_function(parser_function):
        wrapper = _setup_strategy_wrapper(parser_function, strategy)
        return update_wrapper(wrapper, parser_function)

    return decorating_function


def _setup_strategy_wrapper(parser_function, strategy):
    def wrapper(self, *args, **kwargs):
        self.setup_parser_strategy(strategy)
        return parser_function(self, *args, **kwargs)

    return wrapper


class Parser:

    _error_strategies: ClassVar[Dict[str, ErrorStrategy]] = {
        'default': DefaultErrorStrategy(),
        'bail': BailErrorStrategy(),
    }

    def __init__(self, file: str):
        self._parser = self._build_parser(file)
        self._error_listener = self._setup_listener()

    def _setup_listener(self) -> WebIDLErrorListener:
        self._parser.removeErrorListeners()
        self._parser.addErrorListener(error_listener := WebIDLErrorListener())

        return error_listener

    @staticmethod
    def _build_parser(file: str) -> WebIDLParser:
        stream = FileStream(fileName=file, encoding='utf-8')

        functions = (
            WebIDLLexer,
            CommonTokenStream,
            WebIDLParser,
        )

        return cast(
            WebIDLParser,
            reduce(lambda acc, func: func(acc), functions, stream),
        )

    def setup_parser_strategy(self, strategy: str) -> None:
        self._parser._errHandler = self._error_strategies[strategy]

    @_setup_strategy('bail')
    def parse(self) -> ParserRuleContext:
        try:
            return self._parser.webIDL()
        except ParseCancellationException as exc:
            raise IDLParseError from exc

    @_setup_strategy('default')
    def validate(self) -> List[SyntaxErrorInfo]:
        self._parser.webIDL()
        return self._error_listener.errors



from itertools import chain
from pathlib import Path
from typing import Optional, Union

import click
from click import Context, Option, Parameter

from .core import Idl2Js
from .logger import LOG_LEVELS, configure_logging
from .utils import save


def _idl_files(_: Context, __: Union[Option, Parameter], value: Optional[str]) -> tuple[str, ...]:
    if value is None:
        return ()

    return tuple(map(str, Path(value).rglob('*.webidl')))


@click.command()
@click.option(
    '-f', '--file',
    multiple=True,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help='',
)
@click.option(
    '-d', '--directory',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    callback=_idl_files,
    help='',
)
@click.option(
    '-o', '--output',
    type=click.Path(dir_okay=False, resolve_path=True),
    default='/dev/stdout',
    help='',
)
@click.option(
    '-l', '--level',
    default='debug',
    type=click.Choice(LOG_LEVELS.keys()),
    help='',
)
def cli(file: tuple[str, ...], directory: tuple[str, ...], output: str, level: str) -> None:
    configure_logging(level)

    save(
        file_name=output,
        content=Idl2Js(idl=tuple(chain(file, directory))).generate()
    )
